import copy
from paco.models.loader import ModelLoader


def load_project_from_yaml(path, warn=None, validate_local_paths=False):
    """
    Reads a project directory of YAML files and
    returns a complete set of model instances.
    """
    loader = ModelLoader(path, warn, validate_local_paths)
    loader.load_all()
    return loader.project

def deepcopy_except_parent(obj):
    """Returns a deepcopy on on the object, but does not recurse up the
    __parent__ attribute to prevent copying the entire model.
    """
    if not hasattr(obj, '__parent__'):
        return copy.deepcopy(obj)

    # set __parent__ to None, copy the object, then restore __parent__
    parent = getattr(obj, '__parent__', None)
    obj.__parent__ = None
    copy_obj = copy.deepcopy(obj)
    obj.__parent__ = parent
    return copy_obj



from paco.models import schemas
from paco.models import vocabulary
from paco.models.exceptions import InvalidCFNMapping, InvalidAWSResourceName
from paco.models.locations import get_parent_by_interface
from functools import partial
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from paco.models.references import get_model_obj_from_ref
import json
import hashlib
import troposphere
import troposphere.ecs
import zope.schema
import zope.schema.interfaces


def match_allowed_paco_filenames(paco_home, sub_dirname, filename=None):
    """
    Return the Path to a Paco-named directory or file if it can match filename to one a lower-case
    or capitalize case otherwise return None:

    Directory:

      * paco_home / {sub_dirname}
      * paco_home / {Sub_dirname}

    Lower-case sub_dirname:

     * paco_home / {sub_dirname} / {filename}.yaml <-- lower-case
     * paco_home / {sub_dirname} / {Filename}.yaml <-- capitalize
     * paco_home / {sub_dirname} / {filename}.yml <-- .yml extension
     * paco_home / {sub_dirname} / {Filename}.yml <-- .yml extension

    Capitalized sub_dirname:

     * paco_home / {Sub_dirname} / {filename}.yaml <-- lower-case
     * paco_home / {Sub_dirname} / {Filename}.yaml <-- capitalize
     * paco_home / {Sub_dirname} / {filename}.yml <-- .yml extension
     * paco_home / {Sub_dirname} / {Filename}.yml <-- .yml extension

    """
    def filename_match(paco_home, sub_dirname, filename):
        path = paco_home / sub_dirname / f'{filename}.yml'
        if path.is_file():
            return path
        path = paco_home / sub_dirname / f'{filename}.yaml'
        if path.is_file():
            return path
        filename = filename.capitalize()
        path = paco_home / sub_dirname / f'{filename}.yml'
        if path.is_file():
            return path
        path = paco_home / sub_dirname / f'{filename}.yaml'
        if path.is_file():
            return path

    sub_dirname = sub_dirname.lower()
    if filename == None:
        path = paco_home / sub_dirname
        if path.is_dir():
            return path
        sub_dirname = sub_dirname.capitalize()
        path = paco_home / sub_dirname
        if path.is_dir():
            return path
        else:
            return None

    filename = filename.lower()
    path = filename_match(paco_home, sub_dirname, filename)
    if path == None:
        sub_dirname = sub_dirname.capitalize()
        path = filename_match(paco_home, sub_dirname, filename)

    return path

def interface_seen(seen, iface):
    """Return True if interface already is seen.
    """
    for seen_iface in seen:
        if seen_iface.extends(iface):
            return True
    return False

def most_specialized_interfaces(context):
    """Get interfaces for an object without any duplicates.

    Interfaces in a declaration for an object may already have been seen
    because it is also inherited by another interface.
    """
    declaration = zope.interface.providedBy(context)
    seen = []
    for iface in declaration.flattened():
        if interface_seen(seen, iface):
            continue
        seen.append(iface)
    return seen

def marshall_fieldname_to_troposphere_value(obj, props, troposphere_name, field_name):
    """
    Return a value that can be used in a troposphere Properties dict.

    If field_name is a tuple, the first item is the field_name in the schema and the
    second item is the attribute (typically a @property) to return the modified value.
    """

    if type(field_name) == type(tuple()):
        mapping_field_name = field_name[0]
        mapping_attr_name = field_name[1]
    else:
        mapping_field_name = field_name
        mapping_attr_name = field_name

    for interface in most_specialized_interfaces(obj):
        fields = zope.schema.getFields(interface)
        if mapping_field_name in fields:
            # Field provides type information to help convert to Troposphere type
            field = fields[mapping_field_name]
            value = getattr(obj, mapping_attr_name)
            if zope.schema.interfaces.IBool.providedBy(field) and props[troposphere_name][0] == type(str()):
                if value:
                    return 'true'
                else:
                    return 'false'
            if isinstance(field, zope.schema.Text) and props[troposphere_name][0] == type(dict()):
                try:
                    return json.loads(value)
                except TypeError:
                    return None
            if value == {} or value == []:
                return None
            else:
                return value
        elif hasattr(obj, mapping_field_name):
            # supplied as a base property or attribute - these should return
            # types that troposphere expects
            value = getattr(obj, mapping_field_name)
            if value == {} or value == []:
                return None
            else: return value
        else:
            # Shouldn't get here unless the cfn_mapping is incorrect
            raise InvalidCFNMapping(f"cfn_mapping is incorrect for {mapping_field_name} for {interface}?")

def get_all_fields(obj):
    """
    Introspect an object for schemas it implements
    and return a dict of {fieldname: field_obj}
    """
    fields = {}
    try:
        # all interfaces implemented by obj
        interfaces = obj.__provides__.__iro__
    except AttributeError:
        # object without interfaces like a plain dict
        return {}
    for interface in interfaces:
        fields.update(zope.schema.getFields(interface))
    return fields

class CFNExport():
    @property
    def cfn_export_dict(self):
        "CloudFormation export dictionary suitable for use in troposphere templates"
        result = {}
        for key, value in self.cfn_mapping.items():
            value = marshall_fieldname_to_troposphere_value(
                self,
                self.troposphere_props,
                key,
                value
            )
            if value != None:
                result[key] = value
        return result

def md5sum(filename=None, str_data=None):
    """Computes and returns an MD5 sum in hexdigest format on a file or string"""
    d = hashlib.md5()
    if filename != None:
        with open(filename, mode='rb') as f:
            for buf in iter(partial(f.read, 128), b''):
                d.update(buf)
    elif str_data != None:
        d.update(bytearray(str_data, 'utf-8'))
    else:
        raise AttributeError("md5sum: Filename or String data expected")

    return d.hexdigest()

@implementer(schemas.IParent)
class Parent(CFNExport):
    "One base class to rule them all"

    def obj_hash(self):
        fields = get_all_fields(self)
        str_data = ''
        for name, obj in fields.items():
            value = None
            if isinstance(obj, (str, zope.schema.TextLine, zope.schema.Text)):
                value = getattr(self, name)
            if isinstance(obj, (int, float, zope.schema.Int, zope.schema.Float, bool, zope.schema.Bool)):
                value = getattr(self, name)
                value = str(value)

            if value != None:
                str_data += value
                value = None
        hash_data = md5sum(str_data=str_data)
        return hash_data

    def __init__(self, __parent__):
        self.__parent__ = __parent__

    @property
    def paco_ref_list(self):
        "List of paco.ref parts"
        obj = self
        parts = []
        parent = obj.__parent__
        while parent != None:
            name = getattr(obj, 'name', None)
            # objects of type IParent do not have a name, their parent's ref is returned
            if name != None:
                parts.append(name)
            obj = parent
            parent = obj.__parent__
        parts.reverse()
        return parts

    @property
    def paco_ref_parts(self):
        "Bare paco.ref string to the object"
        parts = self.paco_ref_list
        return '.'.join(parts)

    @property
    def paco_ref(self):
        "paco.ref string to the object"
        return 'paco.ref ' + self.paco_ref_parts

@implementer(schemas.ITitle)
class Title():
    title = FieldProperty(schemas.INamed["title"])

@implementer(schemas.INamed)
class Named(Parent):
    """
    An item which has a name and an optional title.
    A name is a unique, human-readable id.

    Every model is location-aware. This means it has a
    __name__ and __parent__. This lets the object live in
    a hiearchy such as a URL tree for a web application, object database
    or filesystem. Paco uses for the web app.
    """
    name = FieldProperty(schemas.INamed["name"])
    title = FieldProperty(schemas.INamed["title"])
    _stack_hooks = None

    def __init__(self, name, __parent__):
        super().__init__(__parent__)
        self.name = name
        self.__name__ = self.name

    @property
    def title_or_name(self):
        "It's a Plone classic!"
        if len(self.title) > 0:
            return self.title
        return self.name

    def add_stack_hooks(self, stack_hooks):
        """
        Add StackHooks to the Resource
        """
        # class attributes are shared, create an instance attribute
        if self._stack_hooks == None:
            self._stack_hooks = []

        # If the Stack has already been initialized, add hooks to it
        # ToDo: duck typing as paco.models can depend upon paco.stack - create Interface for IStack?
        if hasattr(self, 'stack') and hasattr(self.stack, 'add_hooks'):
            self.stack.add_hooks(stack_hooks)
        # Before initialization the hooks are set as a private attribute and added during initialization
        else:
            self._stack_hooks.append(stack_hooks)


@implementer(schemas.IName)
class Name():
    name = FieldProperty(schemas.IName["name"])

@implementer(schemas.IEnablable)
class Enablable():
    """
    Configuration that can be enabled/disabled
    """
    enabled = FieldProperty(schemas.IEnablable["enabled"])

    def is_enabled(self):
        """
        Returns True in a deployed state, otherwise False.
        Will walk up the tree, and if anything is set to "enabled: false"
        this will return False.
        """
        state = self.enabled
        # if current resource is already "enabled: false" simlpy return that
        if not state: return False

        context = self
        while context is not None:
            override = getattr(context, 'enabled', True)
            # once we encounter something higher in the tree with "enabled: false"
            # the lower resource is always False and we return that
            if not override: return False
            context = getattr(context, '__parent__', None)

        # walked right to the top and enabled is still true
        return True

@implementer(schemas.IDeployable)
class Deployable():
    """
    A deployable resource
    """
    enabled = FieldProperty(schemas.IDeployable["enabled"])

    def is_enabled(self):
        """
        Returns True in a deployed state, otherwise False.
        Will walk up the tree, and if anything is set to "enabled: false"
        this will return False.
        """
        state = self.enabled
        # if current resource is already "enabled: false" simlpy return that
        if not state: return False

        context = self
        while context is not None:
            override = getattr(context, 'enabled', True)
            # once we encounter something higher in the tree with "enabled: false"
            # the lower resource is always False and we return that
            if not override: return False
            context = getattr(context, '__parent__', None)

        # walked right to the top and enabled is still true
        return True

@implementer(schemas.INameValuePair)
class NameValuePair(Parent):
    name = FieldProperty(schemas.INameValuePair['name'])
    value = FieldProperty(schemas.INameValuePair['value'])

    troposphere_props = troposphere.ecs.Environment.props
    cfn_mapping = {
        'Name': 'name',
        'Value': 'value'
    }

class Regionalized():
    "Mix-in to allow objects to identify which account and region they are deployed in"

    @property
    def account_name(self):
        account_cont = get_parent_by_interface(self, schemas.IAccountContainer)
        if account_cont != None:
            return account_cont.name
        env_region = get_parent_by_interface(self, schemas.IEnvironmentRegion)
        project = get_parent_by_interface(self)
        if env_region != None:
            return get_model_obj_from_ref(env_region.network.aws_account, project).name
        raise AttributeError('Could not determine account for {}'.format(self.name))

    @property
    def region_name(self):
        # ALlow an object to override it's region_name
        # for example, Route53HealthCheck hard-codes Metrics to us-east-1
        if hasattr(self, 'overrode_region_name'):
            return self.overrode_region_name
        # region the resource is deployed in
        region = get_parent_by_interface(self, schemas.IRegionContainer)
        if region != None:
            return region.name
        # Global buckets have a region field
        if schemas.IS3Bucket.providedBy(self):
            return self.region
        raise AttributeError('Could not determine region for {}'.format(self.name))

    @property
    def region_full_name(self):
        return vocabulary.aws_regions[self.region_name]['full_name']

    @property
    def region_short_name(self):
        return vocabulary.aws_regions[self.region_name]['short_name']

@implementer(schemas.IDNSEnablable)
class DNSEnablable():
    dns_enabled = FieldProperty(schemas.IDNSEnablable['dns_enabled'])
    force_dns_enabled = FieldProperty(schemas.IDNSEnablable['force_dns_enabled'])

    def is_dns_enabled(self):
        """
        Returns True in a deployed state, otherwise False.
        Will walk up the tree, and if anything is set to "enabled: false"
        this will return False.
        """
        if self.force_dns_enabled == True:
            return True
        state = self.dns_enabled
        # if current resource is already "enabled: false" simlpy return that
        if not state: return False

        context = self
        while context is not None:
            override = getattr(context, 'dns_enabled', True)
            # once we encounter something higher in the tree with "enabled: false"
            # the lower resource is always False and we return that
            if not override: return False
            context = getattr(context, '__parent__', None)

        # walked right to the top and enabled is still true
        return True


@implementer(schemas.IImportFrom)
class ImportFrom():
    """
    Configuration that can be enabled/disabled
    """
    import_from = FieldProperty(schemas.IImportFrom["import_from"])

@implementer(schemas.IType)
class Type():
    type = FieldProperty(schemas.IType['type'])

@implementer(schemas.IResource)
class Resource(Type, Named, Deployable, Regionalized, DNSEnablable, ImportFrom):
    "Resource"
    order = FieldProperty(schemas.IResource['order'])
    change_protected = FieldProperty(schemas.IResource['change_protected'])

    def get_account(self):
        """
        Return the Account object that this resource is provisioned to
        """
        region = get_parent_by_interface(self, schemas.IRegionContainer)
        project = get_parent_by_interface(self, schemas.IProject)
        # NetEnv accounts
        if schemas.IEnvironmentRegion.providedBy(region):
            return get_model_obj_from_ref(region.network.aws_account, project)
        # Service accounts
        else:
            account_cont = get_parent_by_interface(self, schemas.IAccountContainer)
            return project.accounts[account_cont.name]

    # Resource Name methods

    def big_join(self, str_list, separator_ch, camel_case=False, none_value_ok=False):
        # Camel Case
        new_str = ""
        first = True
        for str_item in str_list:
            if none_value_ok == True and str_item == None:
                continue
            if first == False:
                new_str += separator_ch
            if camel_case == True:
                new_str += str_item[0].upper()+str_item[1:]
            else:
                new_str += str_item
            first = False
        return new_str

    def resource_name_filter(self, name, filter_id, hash_long_names):
        "Checks a name against a filter and raises an InvalidAWSResourceName if it is not a valid AWS name"
        message = None
        max_name_len = None
        if filter_id in [
            'EC2.ElasticLoadBalancingV2.LoadBalancer.Name',
            'EC2.ElasticLoadBalancingV2.TargetGroup.Name']:
            if len(name) > 32:
                max_name_len = 32
                message = "Name must not be longer than 32 characters.",
            elif filter_id.find('LoadBalancer') != -1 and name.startswith('internal-'):
                message = "Name must not start with 'internal-'"
            elif name[-1] == '-' or name[0] == '-':
                message = "Name must not begin or end with a dash."
        elif filter_id in [
            'IAM.Role.RoleName',
            'IAM.ManagedPolicy.ManagedPolicyName']:
            if len(name) > 255:
                max_name_len = 255
                message = "Name must not be longer than 255 characters."
        elif filter_id == 'IAM.Policy.PolicyName':
            if len(name) > 128:
                max_name_len = 128
                message = "Name must not be longer than 128 characters."
        elif filter_id == 'ElastiCache.ReplicationGroup.ReplicationGroupId':
            if len(name) > 40:
                max_name_len = 255
                message = "ReplicationGroupId must be 40 characters or less"
        elif filter_id == 'SecurityGroup.GroupName':
            pass
        else:
            message = 'Unknown filter_id'

        if max_name_len != None and hash_long_names == True:
            message = None
            name_hash = md5sum(str_data=name)[:8].upper()
            name = name_hash + '-' + name[((max_name_len-9)*-1):]

        if message != None:
            raise InvalidAWSResourceName(
                message="{}: {}: {}".format(
                        filter_id,
                        message,
                        name,
                )
            )
        return name

    def resource_char_filter(self, ch, filter_id, remove_invalids=False):
        # Universal check
        if ch.isalnum() == True:
            return ch
        # SecurityGroup Group Name
        # Constraints for EC2-VPC: a-z, A-Z, 0-9, spaces, and ._-:/()#,@[]+=&;{}!$*
        if filter_id == 'SecurityGroup.GroupName':
            if ch in ' ._-:/()#,@[]+=&;{}!$*':
                return ch
        elif filter_id in [
            'IAM.Role.RoleName',
            'IAM.ManagedPolicy.ManagedPolicyName',
            'IAM.Policy.PolicyName']:
            if ch in '_+=,.@-.':
                return ch
        elif filter_id == 'ElastiCache.ReplicationGroup.ReplicationGroupId':
            if ch in '-':
                return ch
        elif filter_id in [
            'EC2.ElasticLoadBalancingV2.LoadBalancer.Name',
            'EC2.ElasticLoadBalancingV2.TargetGroup.Name']:
            # Only alphanum and dases are allowed
            pass
        else:
            raise StackException(PacoErrorCode.Unknown, message="Invalid filter Id: "+filter_id)

        if remove_invalids == True:
            return ''

        # By default return a '-' for invalid characters
        return '-'

    def create_cfn_logical_id(self, camel_case=False):
        "The logical ID must be alphanumeric (A-Za-z0-9) and unique within the template."
        name = self.type + self.name
        return self.create_resource_name(name, remove_invalids=True, camel_case=camel_case).replace('-', '')

    def create_resource_name(
        self,
        name,
        remove_invalids=False,
        filter_id=None,
        hash_long_names=False,
        camel_case=False):
        """
        Resource names are only alphanumberic (A-Za-z0-9) and dashes.
        Invalid characters are removed or changed into a dash.
        """
        def normalize(name, remove_invalids, filter_id, camel_case):
            uppercase_next_char = False
            new_name = ''
            for ch in name:
                if filter_id != None:
                    ch = self.resource_char_filter(ch, filter_id, remove_invalids)
                    if ch == '' and camel_case == True:
                        uppercase_next_char = True
                elif ch.isalnum() == True:
                    ch = ch
                elif remove_invalids == False:
                    ch = '-'
                elif remove_invalids == True:
                    ch = ''
                    if camel_case == True:
                        uppercase_next_char = True

                if remove_invalids == True and ch != '' and uppercase_next_char == True:
                    new_name += ch.upper()
                    uppercase_next_char = False
                else:
                    new_name += ch
            return new_name

        if name.isalnum() == True:
            new_name = name
        else:
            new_name = normalize(
                name,
                remove_invalids=remove_invalids,
                filter_id=filter_id,
                camel_case=camel_case
            )
        if filter_id != None:
            new_name = self.resource_name_filter(
                new_name,
                filter_id,
                hash_long_names
            )

        return new_name

    def create_resource_name_join(
        self,
        name_list,
        separator,
        camel_case=False,
        filter_id=None,
        hash_long_names=False
    ):
        name = self.big_join(name_list, separator, camel_case)
        return self.create_resource_name(
            name,
            filter_id=filter_id,
            hash_long_names=hash_long_names,
            camel_case=camel_case
        )

class Stack():
    "Unit of provisionable cloud resources"

class HasStack():
    "Resource that has a Stack associated with it"
    stack = None

    def resolve_ref(self, ref):
        return self.stack

@implementer(schemas.IApplicationResource)
class ApplicationResource(Resource):

    @property
    def app_name(self):
        if hasattr(self, '_app_name'):
            return self._app_name
        app = get_parent_by_interface(self, schemas.IApplication)
        self._app_name = app.name
        return self._app_name

    @property
    def group_name(self):
        if hasattr(self, '_group_name'):
            return self._group_name
        group = get_parent_by_interface(self, schemas.IResourceGroup)
        self._group_name = group.name
        return self._group_name

    @property
    def env_name(self):
        if hasattr(self, '_env_name'):
            return self._env_name
        env = get_parent_by_interface(self, schemas.IEnvironment)
        if env == None:
            account = get_parent_by_interface(self, schemas.IAccountContainer)
            self._env_name = account.name
        else:
            self._env_name = env.name
        return self._env_name

    @property
    def env_obj(self):
        if hasattr(self, '_env_obj'):
            return self._env_obj
        env = get_parent_by_interface(self, schemas.IEnvironment)
        if env == None:
            account = get_parent_by_interface(self, schemas.IAccountContainer)
            self._env_obj = account
        else:
            self._env_obj = env
        return self._env_obj

    @property
    def env_region_obj(self):
        if hasattr(self, '_env_region_obj'):
            return self._env_region_obj
        env_region = get_parent_by_interface(self, schemas.IEnvironmentRegion)
        if env_region == None:
            region_cont = get_parent_by_interface(self, schemas.IRegionContainer)
            self._env_region_obj = region_cont
        else:
            self._env_region_obj = env_region
        return self._env_region_obj

    @property
    def netenv_name(self):
        if hasattr(self, '_netenv_name'):
            return self._netenv_name
        netenv = get_parent_by_interface(self, schemas.INetworkEnvironment)
        if netenv == None:
            service = get_parent_by_interface(self, schemas.IService)
            self._netenv_name = service.name
        else:
            self._netenv_name = netenv.name
        return self._netenv_name


@implementer(schemas.IAccountRef)
class AccountRef():
    account = FieldProperty(schemas.IAccountRef['account'])


@implementer(schemas.IAccountContainer)
class AccountContainer(Named, Regionalized, dict):
    pass


@implementer(schemas.IRegionContainer)
class RegionContainer(Named, Regionalized, dict):
    alarm_sets = FieldProperty(schemas.IRegionContainer['alarm_sets'])

@implementer(schemas.IAccountRegions)
class AccountRegions(Parent):
    account = FieldProperty(schemas.IAccountRegions['account'])
    regions = FieldProperty(schemas.IAccountRegions['regions'])



"""
All things IoT
"""

from paco.models import schemas
from paco.models.base import Parent, Named, Resource, ApplicationResource, md5sum
from paco.models.metrics import Monitorable
from paco.models.locations import get_parent_by_interface
from paco.models.formatter import smart_join
from paco.models.exceptions import InvalidModelObject
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty


# IoT Analytics

@implementer(schemas.IStorageRetention)
class StorageRetention():
    expire_events_after_days = FieldProperty(schemas.IStorageRetention['expire_events_after_days'])

@implementer(schemas.IIotAnalyticsStorage)
class IotAnalyticsStorage(Named, StorageRetention):
    bucket = FieldProperty(schemas.IIotAnalyticsStorage['bucket'])
    key_prefix = FieldProperty(schemas.IIotAnalyticsStorage['key_prefix'])

@implementer(schemas.IAttributes)
class Attributes(Named, dict):
    pass

@implementer(schemas.IIoTPipelineActivity)
class IoTPipelineActivity(Named):
    activity_type = FieldProperty(schemas.IIoTPipelineActivity['activity_type'])
    attributes = FieldProperty(schemas.IIoTPipelineActivity['attributes'])
    attribute_list = FieldProperty(schemas.IIoTPipelineActivity['attribute_list'])
    batch_size = FieldProperty(schemas.IIoTPipelineActivity['batch_size'])
    filter = FieldProperty(schemas.IIoTPipelineActivity['filter'])
    function = FieldProperty(schemas.IIoTPipelineActivity['function'])
    math = FieldProperty(schemas.IIoTPipelineActivity['math'])
    thing_name = FieldProperty(schemas.IIoTPipelineActivity['thing_name'])

@implementer(schemas.IIoTPipelineActivities)
class IoTPipelineActivities(Named, dict):
    pass

@implementer(schemas.IDatasetVariable)
class DatasetVariable(Named):
    double_value = FieldProperty(schemas.IDatasetVariable['double_value'])
    output_file_uri_value = FieldProperty(schemas.IDatasetVariable['output_file_uri_value'])
    string_value = FieldProperty(schemas.IDatasetVariable['string_value'])

@implementer(schemas.IDatasetVariables)
class DatasetVariables(Named, dict):
    pass

@implementer(schemas.IDatasetContainerAction)
class DatasetContainerAction(Named):
    image_arn = FieldProperty(schemas.IDatasetContainerAction['image_arn'])
    resource_compute_type = FieldProperty(schemas.IDatasetContainerAction['resource_compute_type'])
    resource_volume_size_gb = FieldProperty(schemas.IDatasetContainerAction['resource_volume_size_gb'])
    variables = FieldProperty(schemas.IDatasetContainerAction['variables'])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.variables = DatasetVariables('variables', self)

@implementer(schemas.IDatasetQueryAction)
class DatasetQueryAction(Named):
    filters = FieldProperty(schemas.IDatasetQueryAction['filters'])
    sql_query = FieldProperty(schemas.IDatasetQueryAction['sql_query'])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.filters = []

@implementer(schemas.IDatasetS3Destination)
class DatasetS3Destination(Named):
    bucket = FieldProperty(schemas.IDatasetS3Destination['bucket'])
    key = FieldProperty(schemas.IDatasetS3Destination['key'])

@implementer(schemas.IDatasetContentDeliveryRule)
class DatasetContentDeliveryRule(Named):
    s3_destination = FieldProperty(schemas.IDatasetContentDeliveryRule['s3_destination'])

@implementer(schemas.IDatasetContentDeliveryRules)
class DatasetContentDeliveryRules(Named, dict):
    pass

@implementer(schemas.IDatasetTrigger)
class DatasetTrigger(Parent):
    schedule_expression = FieldProperty(schemas.IDatasetTrigger['schedule_expression'])
    triggering_dataset = FieldProperty(schemas.IDatasetTrigger['triggering_dataset'])

@implementer(schemas.IIoTDataset)
class IoTDataset(Named, StorageRetention):
    container_action = FieldProperty(schemas.IIoTDataset['container_action'])
    query_action = FieldProperty(schemas.IIoTDataset['query_action'])
    content_delivery_rules = FieldProperty(schemas.IIoTDataset['content_delivery_rules'])
    triggers = FieldProperty(schemas.IIoTDataset['triggers'])
    version_history = FieldProperty(schemas.IIoTDataset['version_history'])

@implementer(schemas.IIoTDatasets)
class IoTDatasets(Named, dict):
    pass

@implementer(schemas.IIoTAnalyticsPipeline)
class IoTAnalyticsPipeline(ApplicationResource, Monitorable):
    channel_storage = FieldProperty(schemas.IIoTAnalyticsPipeline['channel_storage'])
    datastore_name = FieldProperty(schemas.IIoTAnalyticsPipeline['datastore_name'])
    datastore_storage = FieldProperty(schemas.IIoTAnalyticsPipeline['datastore_storage'])
    pipeline_activities = FieldProperty(schemas.IIoTAnalyticsPipeline['pipeline_activities'])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.channel_storage = IotAnalyticsStorage('channel_storage', self)
        self.pipeline_activities = IoTPipelineActivities('pipeline_activities', self)
        self.datastore_storage = IotAnalyticsStorage('dataset_storage', self)
        self.datasets = IoTDatasets('datasets', self)

    def resolve_ref(self, ref):
        return self.stack

@implementer(schemas.IIoTTopicRuleIoTAnalyticsAction)
class IoTTopicRuleIoTAnalyticsAction(Parent):
    pipeline = FieldProperty(schemas.IIoTTopicRuleIoTAnalyticsAction['pipeline'])

# IoT Core

@implementer(schemas.IIoTPolicy)
class IoTPolicy(ApplicationResource):
    policy_json = FieldProperty(schemas.IIoTPolicy['policy_json'])
    variables = FieldProperty(schemas.IIoTPolicy['variables'])

    def get_aws_name(self):
        "Name of the IoT Policy in AWS"
        # NetworkEnvironment or Service name
        name_list = []
        ne = get_parent_by_interface(self, schemas.INetworkEnvironment)
        if ne == None:
            service = get_parent_by_interface(self, schemas.IService)
            if service == None:
                raise InvalidModelObject("""Unable to find an INetworkEnvironment or IService model object.""")
            name_list.append('Service')
            name_list.append(service.name)
        else:
            name_list.append('ne')
            name_list.append(ne.name)

        # Environment name or Blank if one does not exist
        env = get_parent_by_interface(self, schemas.IEnvironment)
        if env != None:
            name_list.append(env.name)
        name_list.append('app')
        app = get_parent_by_interface(self, schemas.IApplication)
        name_list.append(app.name)
        group = get_parent_by_interface(self, schemas.IResourceGroup)
        name_list.append(group.name)
        name_list.append(self.name)
        aws_name = smart_join('-', name_list)
        aws_name = aws_name.replace('_', '-').lower()

        # If the generated policy name is > 128 chars, then prefix a hash of the name
        if len(aws_name) > 128:
            name_hash = md5sum(str_data=aws_name)[:8]
            copy_size = -(128 - 9)
            if aws_name[copy_size] != '-':
                name_hash += '-'
            aws_name = name_hash + aws_name[copy_size:]

        return aws_name


@implementer(schemas.IIoTVariables)
class IoTVariables(Named, dict):
    pass

@implementer(schemas.IIoTTopicRuleLambdaAction)
class IoTTopicRuleLambdaAction(Parent):
    function = FieldProperty(schemas.IIoTTopicRuleLambdaAction['function'])

@implementer(schemas.IIoTTopicRuleAction)
class IoTTopicRuleAction(Parent):
    awslambda = FieldProperty(schemas.IIoTTopicRuleAction['awslambda'])
    iotanalytics = FieldProperty(schemas.IIoTTopicRuleAction['iotanalytics'])

@implementer(schemas.IIoTTopicRule)
class IoTTopicRule(ApplicationResource, Monitorable):
    actions = FieldProperty(schemas.IIoTTopicRule['actions'])
    aws_iot_sql_version = FieldProperty(schemas.IIoTTopicRule['aws_iot_sql_version'])
    rule_enabled = FieldProperty(schemas.IIoTTopicRule['rule_enabled'])
    sql = FieldProperty(schemas.IIoTTopicRule['sql'])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.actions = []

    def resolve_ref(self, ref):
        return self.stack



class UnsupportedFeature(Exception):
    title = "Feature is not yet supported"

class InvalidPacoFieldType(Exception):
    title = "Invalid Paco field type in YAML"

class InvalidPacoSchema(Exception):
    title = "YAML is not a valid Paco schema"

class InvalidPacoProjectFile(Exception):
    title = "Invalid Paco project YAML file"

class UnusedPacoProjectField(Exception):
    title = "Unused Paco project field in YAML"

class InvalidPacoReference(Exception):
    title = "Invalid Paco reference"

class InvalidPacoSub(Exception):
    title = "Invalid Paco subscitition in YAML"

class InvalidCFNMapping(Exception):
    title = "Invalid CloudFormation Mapping from Paco model object"

class InvalidPacoBucket(Exception):
    title = "Invalid Paco Bucket"

class InvalidAWSResourceName(Exception):
    title = "Name is not valid to use as an AWS Resource name"

class InvalidAlarmConfiguration(Exception):
    title = "Alarm configuration is not valid"

class TroposphereConversionError(Exception):
    title = "Troposphere Conversion Error"

class InvalidModelObject(Exception):
    title = "Invalid Model Object"

class InvalidLocalPath(Exception):
    title = "Local path location does not exist"

class AlarmDescriptionExtensionConflict(Exception):
    title = "AlarmDescription extension conflict"

class LoaderRegistrationError(Exception):
    title = "Loader Registration conflict"

class MissingSNSTopics(Exception):
    title = "Missing SNS Topic"

class InvalidSubType(Exception):
    title = "The sub_type for the object is invalid"


"""
IAM: Identity and Access Managment land
"""

from paco.models.base import Enablable, Parent, Named, Deployable
from paco.models import schemas
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from paco.models import loader
import json


@implementer(schemas.IIAM)
class IAM(Named):
    roles = FieldProperty(schemas.IIAM["roles"])
    #policies = FieldProperty(schemas.IIAM["policies"])

    def add_roles(self, roles_config_dict):
        roles_dict = {
            'roles': roles_config_dict
        }
        loader.apply_attributes_from_config(self, roles_dict)

@implementer(schemas.IBaseRole)
class BaseRole(Named):
    instance_profile = FieldProperty(schemas.IBaseRole["instance_profile"])
    path = FieldProperty(schemas.IBaseRole["path"])
    managed_policy_arns = FieldProperty(schemas.IBaseRole["managed_policy_arns"])
    max_session_duration = FieldProperty(schemas.IBaseRole["max_session_duration"])
    permissions_boundary = FieldProperty(schemas.IBaseRole["permissions_boundary"])
    assume_role_policy = FieldProperty(schemas.IBaseRole["assume_role_policy"])
    role_name = FieldProperty(schemas.IBaseRole["role_name"])
    global_role_name = FieldProperty(schemas.IBaseRole["global_role_name"])
    policies = FieldProperty(schemas.IBaseRole["policies"])

    def __init__(self, __name__, __parent__):
        super().__init__(__name__, __parent__)
        self.policies = []
        self.managed_policy_arns = []

    def apply_config(self, config_dict):
        loader.apply_attributes_from_config(self, config_dict)

    def set_assume_role_policy(self, policy_config_dict):
        policy_config = AssumeRolePolicy(self)
        loader.apply_attributes_from_config(policy_config, policy_config_dict)
        self.assume_role_policy = policy_config

    def add_policy(self, policy_config_dict):
        policy_config = Policy(self)
        loader.apply_attributes_from_config(policy_config, policy_config_dict)
        self.policies.append(policy_config)

    def get_arn(self):
        return self.resolve_ref_obj.get_role_arn()

    def resolve_ref(self, ref):
        return ref.resource.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.IRole)
class Role(BaseRole, Deployable):
    pass

@implementer(schemas.IRoleDefaultEnabled)
class RoleDefaultEnabled(BaseRole, Enablable):
    pass

@implementer(schemas.IPolicy)
class Policy(Parent):
    name = FieldProperty(schemas.IPolicy["name"])
    statement = FieldProperty(schemas.IPolicy["statement"])

    def export_as_json(self):
        "Export policy as JSON"
        # Policies are normally included in a CloudFormation template as YAML
        # this JSON export is used by parliment to analyze policy problems.
        policy = {"Version": "2012-10-17", "Statement": []}
        for statement in self.statement:
            policy["Statement"].append(
                {
                    "Effect": statement.effect,
                    "Action": statement.action,
                    "Resource": statement.resource,
                }
            )
        return json.dumps(policy)

@implementer(schemas.IAssumeRolePolicy)
class AssumeRolePolicy(Parent):
    effect = FieldProperty(schemas.IAssumeRolePolicy["effect"])
    aws = FieldProperty(schemas.IAssumeRolePolicy["aws"])
    service = FieldProperty(schemas.IAssumeRolePolicy["service"])

@implementer(schemas.IPrincipal)
class Principal(Named):
    aws = FieldProperty(schemas.IPrincipal["aws"])
    service = FieldProperty(schemas.IPrincipal["service"])


@implementer(schemas.IStatement)
class Statement(Named):
    action = FieldProperty(schemas.IStatement["action"])
    condition = FieldProperty(schemas.IStatement["condition"])
    effect = FieldProperty(schemas.IStatement["effect"])
    resource = FieldProperty(schemas.IStatement["resource"])
    principal = FieldProperty(schemas.IStatement["principal"])

    def __init__(self, __name__, __parent__):
        super().__init__(__name__, __parent__)
        self.action = []
        self.resource = []
        self.condition = {}


@implementer(schemas.IManagedPolicy)
class ManagedPolicy(Named, Deployable):
    policy_name = FieldProperty(schemas.IManagedPolicy["policy_name"])
    statement = FieldProperty(schemas.IManagedPolicy["statement"])
    roles = FieldProperty(schemas.IManagedPolicy["roles"])
    users = FieldProperty(schemas.IManagedPolicy["users"])
    path = FieldProperty(schemas.IManagedPolicy["path"])



from paco.models import schemas


def get_parent_by_interface(context, interface=schemas.IProject):
    """
    Walk up the tree until an object provides the requested Interface
    """
    max = 999
    while context is not None:
        if interface.providedBy(context):
            return context
        if schemas.IProject.providedBy(context):
            return None
        else:
            context = context.__parent__
        max -= 1
        if max < 1:
            raise TypeError("Maximum location depth exceeded. Model is borked!")





import zope.schema
import zope.interface

class InvalidPacoReferenceString(zope.schema.ValidationError):
    __doc__ = 'PacoReference must be of type (string)'

class InvalidPacoReferenceStartsWith(zope.schema.ValidationError):
    __doc__ = "PacoReference must begin with 'paco.ref'"

class InvalidPacoReferenceRefType(zope.schema.ValidationError):
    __doc__ = "PacoReference 'paco.ref must begin with: netenv | resource | accounts | function | service"

class FileReference():
    pass

class StringFileReference(FileReference, zope.schema.Text):
    """Path to a string file on the filesystem"""

    def constraint(self, value):
        """
        Validate that the path resolves to a file on the filesystem
        """
        return True
        # ToDo: how to get the PACO_HOME and change to that directory from here?
        #path = pathlib.Path(value)
        #return path.exists()

class BinaryFileReference(FileReference, zope.schema.Bytes):
    """Path to a binary file on the filesystem"""

    def constraint(self, value):
        """
        Validate that the path resolves to a file on the filesystem
        """
        return True

class YAMLFileReference(FileReference, zope.schema.Object):
    """Path to a YAML file"""

    def __init__(self, **kw):
        self.schema = zope.interface.Interface
        self.validate_invariants = kw.pop('validate_invariants', True)
        super(zope.schema.Object, self).__init__(**kw)

def is_ref(paco_ref, raise_enabled=False):
    """Determines if the string value is a Paco reference"""
    if type(paco_ref) != type(str()):
        if raise_enabled: raise InvalidPacoReferenceString
        return False
    if paco_ref.startswith('paco.ref ') == False:
        if raise_enabled: raise InvalidPacoReferenceStartsWith
        return False
    ref_types = ["netenv", "resource", "accounts", "function", "service"]
    for ref_type in ref_types:
        if paco_ref.startswith('paco.ref %s.' % ref_type):
            return True
    if raise_enabled: raise InvalidPacoReferenceRefType
    return False

class PacoReference(zope.schema.Text):

    def __init__(self, *args, **kwargs):
        self.str_ok = False
        self.schema_constraint = ''
        if 'str_ok' in kwargs.keys():
            self.str_ok = kwargs['str_ok']
            del kwargs['str_ok']
        # schema_constraint is a string name of an ISchema
        # if a Schema is passed, it is converted to a string
        if 'schema_constraint' in kwargs.keys():
            self.schema_constraint = kwargs['schema_constraint']
            if hasattr(self.schema_constraint, '__name__'):
                self.schema_constraint = self.schema_constraint.__name__
            del kwargs['schema_constraint']
        super().__init__(*args, **kwargs)

    def constraint(self, value):
        """
        Limit text to the format 'word.ref chars_here.more-chars.finalchars100'
        """
        if self.str_ok and is_ref(value) == False:
            if isinstance(value, str) == False:
                raise InvalidPacoReferenceString
                #return False
            return True

        return is_ref(value, raise_enabled=True)



"""
Implementation for Logging model objects
"""
from paco.models import schemas
from paco.models.base import Named
from paco.models.locations import get_parent_by_interface
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty


@implementer(schemas.ICloudWatchLogRetention)
class CloudWatchLogRetention():
    expire_events_after_days = FieldProperty(schemas.ICloudWatchLogRetention["expire_events_after_days"])

@implementer(schemas.ICloudWatchLogSources)
class CloudWatchLogSources(Named, dict):
    "Collection of Log Sources"

@implementer(schemas.ICloudWatchLogSource)
class CloudWatchLogSource(Named, CloudWatchLogRetention):
    encoding = FieldProperty(schemas.ICloudWatchLogSource["encoding"])
    log_stream_name = FieldProperty(schemas.ICloudWatchLogSource["log_stream_name"])
    multi_line_start_pattern = FieldProperty(schemas.ICloudWatchLogSource["multi_line_start_pattern"])
    path = FieldProperty(schemas.ICloudWatchLogSource["path"])
    timestamp_format = FieldProperty(schemas.ICloudWatchLogSource["timestamp_format"])
    timezone = FieldProperty(schemas.ICloudWatchLogSource["timezone"])

@implementer(schemas.IMetricTransformation)
class MetricTransformation():
    default_value = FieldProperty(schemas.IMetricTransformation["default_value"])
    metric_name = FieldProperty(schemas.IMetricTransformation["metric_name"])
    metric_namespace = FieldProperty(schemas.IMetricTransformation["metric_namespace"])
    metric_value = FieldProperty(schemas.IMetricTransformation["metric_value"])

@implementer(schemas.IMetricFilters)
class MetricFilters(Named, dict):
    pass

@implementer(schemas.IMetricFilter)
class MetricFilter(Named):
    filter_pattern = FieldProperty(schemas.IMetricFilter["filter_pattern"])
    metric_transformations = FieldProperty(schemas.IMetricFilter["metric_transformations"])

@implementer(schemas.ICloudWatchLogGroups)
class CloudWatchLogGroups(Named, dict):
    pass

@implementer(schemas.ICloudWatchLogGroup)
class CloudWatchLogGroup(Named, CloudWatchLogRetention):
    external_resource = FieldProperty(schemas.ICloudWatchLogGroup["external_resource"])
    metric_filters = FieldProperty(schemas.ICloudWatchLogGroup["metric_filters"])
    sources = FieldProperty(schemas.ICloudWatchLogGroup["sources"])
    log_group_name = FieldProperty(schemas.ICloudWatchLogGroup["log_group_name"])

    def __init__(self, name, __parent__):
        super().__init__(name, __parent__)
        self.metric_filters = MetricFilters('metric_filters', self)
        self.sources = CloudWatchLogSources(name, __parent__)

    def get_log_group_name(self):
        "Return log_group_name if set, otherwise return name"
        if self.log_group_name != '':
            return self.log_group_name
        return self.name

    def get_full_log_group_name(self):
        if self.external_resource == True:
            return self.log_group_name
        name = self.get_log_group_name()
        parent = get_parent_by_interface(self, schemas.ICloudWatchLogSet)
        if parent != None:
            return parent.name + '-' + name
        return name

    def resolve_ref(self, ref):
        return self.resolve_ref_obj.resolve_ref(ref)

@implementer(schemas.ICloudWatchLogSets)
class CloudWatchLogSets(Named, dict):
    "Collection of Log Set objects"

    def get_all_log_sources(self):
        "Return a list of all Log Source in these Log Sets"
        results = []
        for log_set in self.values():
            for log_group in self[log_set.name].log_groups.values():
                for log_source in log_group.sources.values():
                    results.append(log_source)
        return results

    def get_all_log_groups(self):
        "Return a list of all Log Groups in these Log Sets"
        results = []
        for log_set in self.values():
            for log_group in self[log_set.name].log_groups.values():
                results.append(log_group)
        return results

@implementer(schemas.ICloudWatchLogSet)
class CloudWatchLogSet(Named):
    "A set of Log Group objects"
    log_groups = FieldProperty(schemas.ICloudWatchLogSet["log_groups"])

@implementer(schemas.ICloudWatchLogging)
class CloudWatchLogging(Named, CloudWatchLogRetention):
    "Contains global defaults for all CloudWatch logging and log sets"
    log_sets = FieldProperty(schemas.ICloudWatchLogging["log_sets"])
